#!/bin/bash

# This script simply launches your C++ executable and contains 5 entry
# points.

main() {
        DX_APP_WIZARD_NAME
}

scatter() {
        DX_APP_WIZARD_NAME scatter
}

map() {
        DX_APP_WIZARD_NAME map
}

process() {
        DX_APP_WIZARD_NAME process
}

postprocess() {
        DX_APP_WIZARD_NAME postprocess
}
